#if !defined(EGLE_H)
#define EGLE_H

//
// glquake.h - glquake engine datatypes
//
#ifdef WIN32
#include <windows.h>
#include <conio.h>
#include <float.h>
#define ABGR
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <signal.h>
#include <stdarg.h>
#include <time.h>
#include <GL/glut.h>
#include <GL/glu.h>
#include "quake.h"

// for debugging view frustrum culling
//#define CULL_STATS


//
// typedefs and defines
//
#ifdef WIN32
	#define M_PI		3.14159265358979323846
	#define M_PI_2		1.57079632679489661923
#endif

#define EGLE_VERSION	"Egle 0.5"

#if !defined(FALSE)
    #define FALSE 0
    #define TRUE 1
#endif

// 60Hz
#define SIM_TICK			(1000/60)

#define SPEEDUP				0.5

#define MOTION_DECAY		0.6
#define STOPPED_LIMIT	    0.01

#define STEP_HEIGHT			8.0
#define HIP_FRACTION		0.75
#define FLY_DIST		    10.0

#define CLIPPING_EPSILON    0.01

#define RUNNING_ACCEL	    (6.0 * SPEEDUP)
#define RUNNING_TURN_SPEED  (12.0 * SPEEDUP)
#define WALKING_ACCEL	    (3.0 * SPEEDUP)
#define WALKING_TURN_SPEED  (6.0 * SPEEDUP)
#define JUMP_ACCEL			(40.0 * SPEEDUP)
#define STEP_SPEED_SCALE    (0.01 * SPEEDUP)
#define CENTER_SPEED	    (WALKING_ACCEL * 0.5)

#define GRAVITY				(4.0 * SPEEDUP)

typedef struct
{
	int			Type;		// object type
	float		fHeight;	// height
	float		fRadius;	// radius of object
	float		fMass;		// mass of object
	vec3_t		Head;		// position of head, used for player views
	vec3_t		View;		// normal along viewing direction
	vec3_t		Feet;		// position of feet
	float		fHeading;	// direction of travel
	float		fPitch;		// neck angle
	float		fRoll;		// not used yet
	float		fSpeed;		// speed along fHeading
	float		fStrafeSpeed;// speed perpendicular to fHeading
	float		fVertSpeed;	// up/down speed
	float		fStep;		// counter for walking
	int			bGravity;	// gravity enabled
	int			bOnGround;	// feet on solid ground
	int			bInLiquid;	// head suspended in liquid (fall linearly)
	int			bCrushed;	// head to feet distance reduced
} OBJECT, *POBJECT;

// must be a power of two
#define ANIM_TEX_FRAMES		32
#define ANIM_TEX_MASK		(ANIM_TEX_FRAMES - 1)
#define ANIM_TEX_DISP		0.1
#define AnimTexIndex( x )   (((x) >> 3) & ANIM_TEX_MASK)

typedef struct
{
    vec3_t		S;
    vec3_t		E;
    int			nSplitNode;
    int			nNode;
} CLIPSTACK, *PCLIPSTACK;

#define CLIPPLANE_END	(-1)
#define CLIPSTACK_DEPTH	256
#define CLIPSEG_LIMIT	256

#define CopyVec3( d, s ) \
( \
    (((d)[0]) = ((s)[0])), \
    (((d)[1]) = ((s)[1])), \
    (((d)[2]) = ((s)[2])) \
)

#define Vec3OpVec3( d, a, op, b ) \
( \
    (((d)[0]) = ((a)[0]) op ((b)[0])), \
    (((d)[1]) = ((a)[1]) op ((b)[1])), \
    (((d)[2]) = ((a)[2]) op ((b)[2])) \
)

#define Vec3OpScl( d, a, op, b ) \
( \
    (((d)[0]) = ((a)[0]) op (b)), \
    (((d)[1]) = ((a)[1]) op (b)), \
    (((d)[2]) = ((a)[2]) op (b)) \
)

#define DotVec3( a, b ) \
( \
    ((a)[0]) * ((b)[0]) + \
    ((a)[1]) * ((b)[1]) + \
    ((a)[2]) * ((b)[2]) \
)

#define CLIP_EPSILON 0.01

#define Notify(x)	if (bVerbose) NULL; else printf x, fflush( stdout )
#define Progress(x)	if (!bVerbose) { printf( "." ); fflush( stdout ); } else printf x
#define ProgressItem(x)	if (!bVerbose) NULL; else printf x 
#define BAIL(x)		{ printf x; goto Abort; }

// primitive offsets and counts for surfaces (lame roll-your-own dlists to keep impact happy)
typedef struct
{
    unsigned int	uOffset : 24;	// offset in pSVArray
    unsigned int	uCount : 8;		// count of primitives
} DLISTDESC, *PDLISTDESC;

// primitive descriptor for lame roll-your-own dlists
typedef union
{
    struct
    {
	short int		sMode;			// usually GL_TRIANGLE_FAN
	short int		sCount;			// consecutive verticies following this entry
    } p;
    struct
    {
	vec2_t		t;					// texture coordinate
	vec2_t		lm[MAXLIGHTBLENDS];	// lightmap texture coordinate
	vec3_t		v;					// vertex coordinate
    } c;
} PRIMDESC, *PPRIMDESC;

// View structure
typedef struct {
	int				ScrX, ScrY;		// Bottem-left corner of view
	int				Width, Height;	// Width and height of the window
	POBJECT			pObject;		// Object which carries the camera
	int				ObjHeight;		// Height above object
	float			HAngle;			// Direction relative to obj dir
	vec3_t			View;			// normal along viewing direction
} VIEW, *PVIEW;

// this is the big structure that describes and episode and map
typedef struct
{
    dheader_t		Head;			// bsp header from PAK file
    int				nV;				// number of verticies
    vec3_t			*pV;			// array of verticies
    int				nE;				// number of edges
    edge_t			*pE;			// array of edges
    int				nEL;			// number of edge lists
    long			*pEL;			// array of edgelists
    int				nP;				// number of planes
    plane_t			*pP;			// array of planes
    int				nTG;			// number of texgen descriptors
    texgen_t		*pTG;			// array of texgens
    int				nS;				// number of surfaces
    int				nMaxEdgesPerSurface;// most edges in a surface
    surface_t		*pS;			// array of surfaces
    unsigned short int *pSortedS;	// array of surface indicies sorted by texgen
    unsigned short int usTessSurf;	// index of currently tesselating surface
    PDLISTDESC		pSDL;			// surface display list descriptor
    unsigned int	uSPAMax;		// maximum size of vertex array
    unsigned int	uNextSPA;		// next free prim/vertex slot (for tesselator callback)
    unsigned int	uThisPrim;		// current primitive (for tesselator callback)
    PPRIMDESC		pSPA;			// surface primitive array
    int				nLS;			// number of list surfaces
    unsigned short int *pLS;		// array of list surfaces
    unsigned short int *pTmpLS;		// temp workspace w/ [nLS] elements
    int				nVL;			// size of vislists
    unsigned char	*pVL;			// pointer to vislists in mapped file
    int				nH;				// number of hulls (world is 0)
    bsphull_t		*pH;			// array of hulls
    int				nHN;			// number of bounding hull nodes
    bsphullnode_t	*pHN;			// array of bounding hull nodes
    int				nN;				// number of nodes
    bspnode_t		*pN;			// array of nodes
    int				nL;				// number of leaves
    bspleaf_t		*pL;			// array of leaves
    int				nLM;			// size of lightmaps
    unsigned char	*pLM;			// pointer to lightmaps in mapped file
    unsigned int	uLightSize[2];	// W x H of lightmap texture
    GLuint			uLightObj;		// lightmap texture object
    mipheader_t		*pMipHead;		// mip header
    miptex_t		*pMipTex;		// array of mip-mapped textures (converted to texobjs)
    unsigned char	**ppucTexture;	// pointer to raw 8-bit textures in mapped file
    GLuint			*puTexObj;		// array of texture objects from bsp file
    unsigned int	*puTexObjBase;	// indicies in puTexObj for each miptex's converted texobj (base of anim sequences)
    GLfloat			M[16];			// texgen/lightmap working matrix
    GLfloat			*pfTGM;			// two columns of the texgen matricies
    unsigned char	bLightOn[MAXLIGHTSTYLES];// light status
    unsigned char	ucLightUsed[MAXLIGHTSTYLES];// array of light styles in use
    int				nLightStyles;	// number of light styles in use
	PVIEW			pView[4];		// pointer to array of views
} EPISODE, *PEPISODE;

//
// globals
//

extern int				W, H;			// window width and height
extern int				bVerbose;		// how much info goes to std out
extern int				nPixelFormat;	// screen pixal format, should be GL_RGBA
extern int				nTexelFormat;	// number of texture components, should be 4
extern double			dRealTime;		// realtime in milliseconds
extern double			dSimTime;		//
extern unsigned int		uFrame;
extern double			fps;
extern unsigned int		uElapsedFrames;
extern char				*pszMap;
extern char				*pGLVendor;
extern char				*pGLRenderer;


extern unsigned int		Palette[256];
	
extern PEPISODE			pCurrEpsd;

extern OBJECT			Player;
extern int				bDrawHulls;
extern int				bDrawTexture;

extern int				usHull;
// I suspect 1 is the shot clipping hull, 2 is the walking clipping hull...
extern int				usClipHull;


//
// prototypes
//
int CmpSurfTex( const void *p1, const void *p2 );
void RenderLevel( void );
int IsPointOutsideHull( int nNode, vec3_t *p );
int BuildTextures( void );
int BuildLightmaps( void );
int CalcTexCoords( void );
int LoadFiles( void );
PEPISODE CreateEpisode( void );
int DestroyEpisode( PEPISODE pEpsd );
int LoadMap( char * pszName );


#endif /* !defined(EGLE_H) */
