#if !defined(QUAKE_H)
#define QUAKE_H

//
// quake.h - quake datatypes
//

#if !defined(FALSE)
    #define FALSE 0
    #define TRUE 1
#endif

typedef unsigned char byte;

#define QUAKE_MAGIC "PACK"

// main directory pointer (at 0 in .pak)
typedef struct 
{ 
    byte magic[4];		// QUAKE_MAGIC
    long diroffset;             // Position of WAD directory from start of file
    long dirsize;               // Number of entries * 0x40 (64 char)
} pakheader_t;

// dir entries
typedef struct 
{ 
    byte filename[0x38];      // Name of the file, Unix style, with extension
    long offset;                // Position of the entry in PACK file
    long size;                  // Size of the entry in PACK file
} pakentry_t;

// BSP entries
typedef struct
{
    long  offset;               // Offset to entry, in bytes, from start of file
    long  size;                 // Size of entry in file, in bytes
} dentry_t;

// Here is the BSP header itself, made of a version tag, and 14 entries: 

#define BSP_VERSION 29

// The BSP file header
typedef struct
{
    long  version;              // Model version, must be 27.
    dentry_t entities;          // List of Entities.
    dentry_t planes;            // Map Planes
    dentry_t miptex;            // Wall Textures.
    dentry_t vertices;          // Map Vertices.
    dentry_t visilist;		// Leaves Visibility lists.
    dentry_t nodes;             // BSP Nodes.
    dentry_t texgen;		// Texture coordinate generator.
    dentry_t surfaces;          // Map Surfaces.
    dentry_t lightmaps;         // Wall Light Maps.
    dentry_t boundnodes;	// BSP to bound Hulls.
    dentry_t leaves;            // Hull BSP Leaves.
    dentry_t lstsurf;           // List of Surfaces.
    dentry_t edges;             // Original surface Edges.
    dentry_t lstedges;          // List of surfaces Edges.
    dentry_t hulls;             // List of Hulls.
} dheader_t;

typedef float scalar_t;		// scalar value
typedef float vec2_t[3];	// S,T coordinates of the vertex
typedef float vec3_t[3];	// X,Y,Z coordinates of the vertex
typedef short svec3_t[3];	// X,Y,Z coordinates of the vertex in shorts

typedef struct
{
    vec3_t min;			// minimum values of X,Y,Z
    vec3_t max;			// maximum values of X,Y,Z
} boundbox_t;

typedef struct
{
    svec3_t min;		// minimum values of X,Y,Z in shorts
    svec3_t max;		// maximum values of X,Y,Z in shorts
} sboundbox_t;

typedef struct
{
    u_short startvertex;	// id of the start vertex
    u_short endvertex;		// id of the start vertex
} edge_t;

// Mip texture list header
typedef struct
{
    long numtex;		// Number of textures in Mip Texture list
    long offset[1];		// [numtex] Offsets to each of the individual texture from the begining of mipheader_t
} mipheader_t;

// Mip Texture
typedef struct
{
    char name[16];              // Name of the texture.
    u_long width;               // width of picture, must be a multiple of 8
    u_long height;              // height of picture, must be a multiple of 8
    u_long offset1;             // -> byte Pix[width   * height]
    u_long offset2;             // -> byte Pix[width/2 * height/2]
    u_long offset4;             // -> byte Pix[width/4 * height/4]
    u_long offset8;             // -> byte Pix[width/8 * height/8]
} miptex_t;

#define WAD2_MAGIC "WAD2"

typedef struct 
{ 
    byte magic[4];            // "WAD2", Name of the new WAD format
    long numentries;            // Number of entries
    long diroffset;             // Position of WAD directory in file
} wadhead_t;

typedef struct
{
    long offset;                // Position of the entry in WAD
    long dsize;                 // Size of the entry in WAD file
    long size;                  // Size of the entry in memory (???)
    char type;                  // type of entry
    char cmprs;                 // Compression. 0 if none.
    short dummy;                // Not used
    char name[16];              // 1 to 16 characters, '\0'-padded
} wadentry_t;

#define WAD2_TYPE_PALETTE	'@'
#define WAD2_TYPE_STATPICT	'B'
#define WAD2_TYPE_MIPTEXTURE	'D'
#define WAD2_TYPE_CONSPICT	'E'

// 0-2 are axial planes
#define	PLANE_X			0
#define	PLANE_Y			1
#define	PLANE_Z			2
// 3-5 are non-axial planes snapped to the nearest
#define	PLANE_ANYX		3
#define	PLANE_ANYY		4
#define	PLANE_ANYZ		5

typedef struct
{
    vec3_t normal;		// Vector orthogonal to plane (Nx,Ny,Nz) with Nx+Ny+Nz = 1
    scalar_t dist;		// Offset to plane, along the normal vector
    long type;			// plane type
} plane_t;

// this seems to be a texture coordinate generator
typedef struct
{
    vec3_t snrm;		// s projection (always axis-aligned normal?)
    scalar_t soff;		// s offset (always integral?)
    vec3_t tnrm;		// t projection (always axis-aligned normal?)
    scalar_t toff;		// t offset (always integral?)
    long texid;			// seems to be a texture ID (0-67 in e1m1, has repeats)
    long flag;			// flag for sky/water texture (1 => special texcoords)
} texgen_t;

#define MAXLIGHTMAPS		4
#define MAXLIGHTBLENDS		(1 << MAXLIGHTMAPS)
#define MAXLIGHTSTYLES		256

#define MAXMODELS		256

typedef struct
{
    u_short planenum;		// The plane in which the surface lies
    u_short side;               // 0 if in front of the plane, 1 if behind the plane
    long firstedge;		// first edge in the List of edges
    u_short numedge;		// number of edges in the List of edges
    u_short texgenid;		// index into texgen table
    byte lightstyle[MAXLIGHTMAPS]; // type of light or 0xFF for no light map
    long lightmap;		// offset to the lightmap or -1
} surface_t;

typedef struct
{
    boundbox_t box;		// The bounding box of the Hull
    vec3_t origin;		// always zero; set from entities?
    long headnode[4];		// heads of BSP hulls (0 == main rendering hull?)
    long visleaves;		// number of visible BSP leaves
    long firstsurf;		// id of Surfaces
    long numsurf;		// number of Surfaces
} bsphull_t;

typedef struct
{
    long planenum;		// The plane which splits the node
    short front;		// If positive, id of Front child node, -2 => front inside, -1 => front outside
    short back;			// If positive, id of Back child node, -2 => back inside, -1 => back outside
} bsphullnode_t;

#define CLIP_OUTSIDE	(-1)
#define CLIP_INSIDE	(-2)
// used by BSP clipper
#define NOCLIP_OUTSIDE	(-3)
#define NOCLIP_INSIDE	(-4)

typedef struct
{
    long planenum;		// the plane that splits the node, must be in [0,numplanes)
    short front;		// index to (front >= 0) ? front child node : leaf[~front]
    short back;			// index to (back >= 0) ? back child node : leaf[~back]
    sboundbox_t sbox;		// bounding box (for sphere culling?)
    u_short firstsurf;		// first surface
    u_short numsurf;		// number of surfaces
} bspnode_t;

typedef struct
{
    long type;			// Special type of leaf
    long vislist;		// Beginning of visibility lists, must be -1 or in [0,numvislist)
    sboundbox_t sbox;		// bounding box (usually)
    short firstsurf;		// firstsurf in the lstsurf
    short numsurf;		// number of entries in lstsurf
    byte ambientsnd[4];	// ambient sound volumes, ff = max (4 ambient sounds per level?)
} bspleaf_t;

#define	BSPLEAF_AMBIENT_WATER	0 // watery sound
#define	BSPLEAF_AMBIENT_SKY	1 // wind
#define	BSPLEAF_AMBIENT_SLIME	2 // acid pool
#define	BSPLEAF_AMBIENT_LAVA	3 // for that burning feeling

#define BSPLEAF_TYPE_NORMAL	-1 // ordinary leaf, nothing happens 
#define BSPLEAF_TYPE_INVALID	-2 // the leaf is invalid, so the level is not displayed. 
#define BSPLEAF_TYPE_WATER	-3 // Water, the vision is troubled. 
#define BSPLEAF_TYPE_SLIME	-4 // Acid, like water but green, and hurts the player. 
#define BSPLEAF_TYPE_LAVA	-5 // Lava, vision turns red and the player is badly hurt. 
#define BSPLEAF_TYPE_SKY	-6 // Behaves like water, but is used for sky. 
// All types below -6 seem to be water (maybe they are not implemented yet)

// alias model

#define MDL_MAGIC "IDPO"
#define MDL_VERSION 6

typedef struct
{
    byte magic[4];		// MDL_MAGIC
    long version;		// Version = 6
    vec3_t scale;		// Model scale factors.
    vec3_t origin;		// Model origin.
    scalar_t radius;		// Model bounding radius.
    vec3_t offset;		// Eye position (usefull for parading as a monster)
    long numskins;		// the number of skin textures
    long skinwidth;		// Width of skin texture (must be multiple of 8)
    long skinheight;		// Height of skin texture (must be multiple of 8)
    long numverts;		// Number of vertices
    long numtris;		// Number of triangles surfaces
    long numframes;		// Number of frames
    long synctype;		// 0 = synchron, 1 = random (?)
    long flags;			// 0 (see Alias models)
    scalar_t size;		// average size of triangles
} mdl_t;

typedef struct
{
    long onseam;		// 0 or 0x20, (0x20 && !facesfront) => s += skinwidth / 2
    long s;			// position horizontally, in range [0,skinwidth)
    long t;			// position vertically, in range [0,skinheight)
} mdl_texgen_t;

typedef struct
{
    long facesfront;		// boolean
    long vertidx[3];		// Index of 3 triangle vertices in range [0,numverts)
} mdl_triangle_t;

typedef struct
{
    u_char packedposition[3];	// X,Y,Z coordinate, packed on 0-255
    u_char normalidx;		// index of the lighting vertex normal in [0,162)
} mdl_vert_t;

typedef struct
{
    mdl_vert_t min;		// minimum values of X,Y,Z
    mdl_vert_t max;		// maximum values of X,Y,Z
    char name[16];		// name of frame
    mdl_vert_t v[1];		// [numverts] vertices
} mdl_simpleframe_t;



// this file is shared by quake and qcc

typedef int	func_t;
typedef int	string_t;

typedef enum {ev_void, ev_string, ev_float, ev_vector, ev_entity, ev_field, ev_function, ev_pointer} etype_t;


#define	OFS_NULL		0
#define	OFS_RETURN		1
#define	OFS_PARM0		4		// leave 3 ofs for each parm to hold vectors
#define	OFS_PARM1		7
#define	OFS_PARM2		10
#define	OFS_PARM3		13
#define	OFS_PARM4		16
#define	OFS_PARM5		19
#define	OFS_PARM6		22
#define	OFS_PARM7		25
#define	RESERVED_OFS	28


enum {
	OP_DONE,
	OP_MUL_F,
	OP_MUL_V,
	OP_MUL_FV,
	OP_MUL_VF,
	OP_DIV_F,
	OP_ADD_F,
	OP_ADD_V,
	OP_SUB_F,
	OP_SUB_V,
	
	OP_EQ_F,
	OP_EQ_V,
	OP_EQ_S,
	OP_EQ_E,
	OP_EQ_FNC,
	
	OP_NE_F,
	OP_NE_V,
	OP_NE_S,
	OP_NE_E,
	OP_NE_FNC,
	
	OP_LE,
	OP_GE,
	OP_LT,
	OP_GT,

	OP_LOAD_F,
	OP_LOAD_V,
	OP_LOAD_S,
	OP_LOAD_ENT,
	OP_LOAD_FLD,
	OP_LOAD_FNC,

	OP_ADDRESS,

	OP_STORE_F,
	OP_STORE_V,
	OP_STORE_S,
	OP_STORE_ENT,
	OP_STORE_FLD,
	OP_STORE_FNC,

	OP_STOREP_F,
	OP_STOREP_V,
	OP_STOREP_S,
	OP_STOREP_ENT,
	OP_STOREP_FLD,
	OP_STOREP_FNC,

	OP_RETURN,
	OP_NOT_F,
	OP_NOT_V,
	OP_NOT_S,
	OP_NOT_ENT,
	OP_NOT_FNC,
	OP_IF,
	OP_IFNOT,
	OP_CALL0,
	OP_CALL1,
	OP_CALL2,
	OP_CALL3,
	OP_CALL4,
	OP_CALL5,
	OP_CALL6,
	OP_CALL7,
	OP_CALL8,
	OP_STATE,
	OP_GOTO,
	OP_AND,
	OP_OR,
	
	OP_BITAND,
	OP_BITOR
};


typedef struct statement_s
{
	unsigned short	op;
	short	a,b,c;
} dstatement_t;

typedef struct
{
	unsigned short	type;		// if DEF_SAVEGLOBGAL bit is set
								// the variable needs to be saved in savegames
	unsigned short	ofs;
	int			s_name;
} ddef_t;
#define	DEF_SAVEGLOBGAL	(1<<15)

#define	MAX_PARMS	8

typedef struct
{
	int		first_statement;	// negative numbers are builtins
	int		parm_start;
	int		locals;				// total ints of parms + locals
	
	int		profile;		// runtime
	
	int		s_name;
	int		s_file;			// source file defined in
	
	int		numparms;
	byte	parm_size[MAX_PARMS];
} dfunction_t;


#define	PROG_VERSION	6
typedef struct
{
	int		version;
	int		crc;			// check of header file
	
	int		ofs_statements;
	int		numstatements;	// statement 0 is an error

	int		ofs_globaldefs;
	int		numglobaldefs;
	
	int		ofs_fielddefs;
	int		numfielddefs;
	
	int		ofs_functions;
	int		numfunctions;	// function 0 is an empty
	
	int		ofs_strings;
	int		numstrings;		// first string is a null string

	int		ofs_globals;
	int		numglobals;
	
	int		entityfields;
} dprograms_t;


#endif // !defined(QUAKE_H)
